module Stargate
  module Client
    class Protocol
      UnsupportedProtocolError = Class.new(::Stargate::Client::Error)

      include Logging
      include Stargate::Marshal

      DEFAULT_TIMEOUT = 15

      @@transports = {}

      # Public: List of registered transports (protocol implementations).
      def self.transports
        @@transports
      end

      # Public: Registers new implementation for given protocols.
      def self.register(protocols, transport)
        protocols.each { |protocol| transports[protocol] = transport }
      end

      # Public: Finds implementation for given protocol.
      #
      # Returns transport if found.
      # Raises UnsupportedProtocolError if not found.
      def self.[](protocol)
        transports[protocol.downcase] or raise UnsupportedProtocolError, "Unsupported protocol: #{protocol}"
      end

      # Public: Codec used to serialize/deserialize payload.
      attr_reader :codec

      # Public: Remote URL of the endpoint.
      attr_reader :uri

      # Public: URL without credentials part.
      attr_reader :safe_uri

      # Public: Transport configuration.
      attr_reader :options

      # Public: Constructor.
      def initialize(codec, uri)
        @codec        = codec
        @uri          = uri
        @safe_uri     = uri.dup.tap { |safe_uri| safe_uri.user, safe_uri.password = nil, nil }
        @options      = default_options.merge(parse_options_from_uri_fragment)
        @unmarshaller = Unmarshaller.new
      end

      # Public: Shall load all definitions from remote server.
      def fetch_definitions
        raise NotImplementedError
      end

      # Public: Shall execute class method with given arguments. Executes method remotely of course.
      def call(klass, method, *args)
        raise NotImplementedError
      end

      protected

      def default_options
        { timeout: DEFAULT_TIMEOUT }
      end

      # Internal: A helper to unpack loaded data.
      def unpack_payload(data)
        @unmarshaller.unmarshal(data)
      end

      # Internal: A helper to unpack remote definitions.
      def unpack_definitions(definitions)
        definitions.map { |metadata| ::Stargate::Metadata.from_hash(metadata) }
      end

      # Internal: Parses options from URI fragment. Example http+json://example.com/v1#timeout=0.3
      #
      # Returns hash with options.
      def parse_options_from_uri_fragment
        uri.fragment.to_s.split(';').inject({}) do |options, entry|
          k, v = entry.split('=')
          options[k.to_sym] = v
          options
        end
      end
    end

    require 'stargate/client/protocol/http'
    require 'stargate/client/protocol/inproc'
  end
end
