module Stargate
  module Marshal
    # Internal: Use it in order to pack data for exchange. Marshaller is created per instance of registry
    # version. It packs basic values and objects of registered classes into Stargate::Marshal::Payload wraps.
    class Marshaller
      # Public: Constructor. Initializes marshaller for given registry version object.
      def initialize(registry_version)
        @registry_version = registry_version
      end

      # Public: Returns given object after marshalling.
      def marshal(obj)
        case obj
        when String, Numeric, TrueClass, FalseClass, Float, BigDecimal, NilClass
          Payload.new(Payload::SIMPLE, obj)
        when Symbol
          Payload.new(Payload::SYMBOL, obj.to_s)
        when Hash
          Payload.new(Payload::HASH, marshal_hash(obj))
        when Array
          Payload.new(Payload::LIST, marshal_list(obj))
        else
          Payload.new(*marshal_complex_object(obj))
        end
      end

      private

      # Internal: Returns array marshalled.
      def marshal_list(obj)
        obj.map { |x| marshal(x) }
      end

      # Internal: Returns hash marshalled.
      def marshal_hash(obj)
        obj.inject({}) do |res,(k,v)|
          res[k.to_s] = marshal(v)
          res
        end
      end

      # Internal: Checks if type of given object is registered for exchange. No? Raises an error. Yes? Marshals
      # all public attributes and computed readers.
      #
      # Returns marshalled object.
      def marshal_complex_object(obj)
        metadata = @registry_version.metadata_for(obj)

        attributes = (metadata.attributes + metadata.readers).inject({}) do |result,key|
          result[key.to_s] = marshal(obj.send(key.to_sym))
          result
        end

        [ metadata.name, attributes ]
      end
    end
  end
end
