module Stargate
  # Internal: We need a way to store information about registered classes. Each served class
  # defines itself with the following information:
  #
  # * Class name (eventually serving alias)
  # * List of exposed class methods
  # * List of instance attributes (accessors)
  # * List of instance methods (cacheable reader methods)
  #
  class Metadata
    include Serialization

    # Internal: Loads metadata information from hash.
    def self.from_hash(hash)
      hash.symbolize_keys!

      new(hash[:klass], hash[:name]).tap do |metadata|
        metadata.class_methods(*hash.fetch(:class_methods, []))
        metadata.attributes(*hash.fetch(:attributes, []))
        metadata.readers(*hash.fetch(:readers, []))
      end
    end

    # The class to be served and it's serving alias.
    attr_reader :klass, :name

    # Public: Constructor.
    def initialize(klass, name, &block)
      @klass, @name = klass, name
      @class_methods = []
      @attributes = []
      @readers = []

      instance_eval(&block) if block_given?
    end

    def class_methods(*names)
      load_or_add(:class_methods, *names)
    end

    def attributes(*names)
      load_or_add(:attributes, *names)
    end

    def readers(*names)
      load_or_add(:readers, *names)
    end

    def serialize
      {
        name: name,
        class_methods: class_methods,
        attributes: attributes,
        readers: readers
      }
    end

    def inspect
      "#<#{self.class.name} name=#{name.inspect} class_methods=#{class_methods.inspect} attributes=#{attributes.inspect} readers=#{readers.inspect}"
    end

    private

    # Internal: A helper to define single access methods (getter/setter DSL methods).
    def load_or_add(method, *args)
      value = instance_variable_get("@#{method}")
      args.empty? ? value : value.concat(args.flatten.map(&:to_sym))
    end
  end
end
